(in-package :sockets)

;;;; Methods and other supporting functions for INET-SOCKETs

;;; Our class and constructor

(defclass inet-socket (socket))

(defun make-inet-socket (type protocol) ;exported
  "Make an INET socket.  TYPE may be :datagram or :stream and PROTOCOL
is a protocol number - see GET-PROTOCOL-BY-NAME"
  (let ((fd (sockint::socket sockint::AF-INET
                             (cadr (assoc type `((:datagram
                                                  ,sockint::sock-dgram)
                                                 (:stream
                                                  ,sockint::sock-stream))))
                             protocol)))
    (if (= fd -1)
        (socket-error "socket") 
      (let ((s (make-instance 'inet-socket :descriptor fd)))
        (ext:finalize s (lambda () (socket-close s)))))))
         



;;; binding a socket to an address and port

(defun make-inet-address (dotted-quads)
  "Return a vector of octets given a string DOTTED-QUADS in the format
\"127.0.0.1\""
  (coerce
   (mapcar #'parse-integer
           (split dotted-quads nil '(#\.)))
   'vector))

(defconstant inet-address-any (vector 0 0 0 0))

;; we deliberately redesign the "bind" interface: instead of passing a
;; sockaddr_somethng as second arg, we pass the elements of one as
;; multiple arguments.

;; this is done on the basis that it should make the sockaddr_* stuff
;; all go away.  I hope I'm right.

(defmethod socket-bind ((socket inet-socket)
                        &rest address
                        &aux (host (first address)) (port (second address)))
  (declare (optimize (debug 3)))
  "Bind SOCKET to HOST:PORT.  HOST is a vector of bytes as returned
by MAKE-INET-ADDRESS, and PORT is (INTEGER 0 65535).  See also bind(2)"
  (let* ((sockaddr (sockint::allocate-sockaddr-in))
         (s-port (sockint::sockaddr-in-port sockaddr))
         (s-addr (sockint::sockaddr-in-addr sockaddr)))
    ;; XXX need unwind-protect to free sockaddr
    (setf (sockint::sockaddr-in-family sockaddr) sockint::af-inet)

    ;; port and host are represented in C as "network-endian" unsigned
    ;; integers of various lengths.  This is stupid.  The value of the
    ;; integer doesn't matter (and will change depending on your
    ;; machine's endianness); what the bind(2) call is interested in
    ;; is the pattern of bytes within that integer.

    ;; We have no truck with such dreadful type punning.  We treat
    ;; bytes as bytes

    ;; port
    (setf (alien::deref s-port 0) (ldb (byte 8 8) port))
    (setf (alien::deref s-port 1) (ldb (byte 8 0) port))

    ;; addr
    (setf (alien::deref s-addr 0) (elt host 0))
    (setf (alien::deref s-addr 1) (elt host 1))
    (setf (alien::deref s-addr 2) (elt host 2))
    (setf (alien::deref s-addr 3) (elt host 3))
        
    (if (= (sockint::bind (socket-file-descriptor socket)
                          sockaddr sockint::size-of-sockaddr-in)
           -1)
        (socket-error "bind"))))



(defmethod socket-accept ((socket inet-socket))
  "Perform the accept(2) call, returning a newly-created connected socket,
an inet-address, and a port as multiple values"
  ;;; XXX free sockaddr
  (let* ((sockaddr (sockint::allocate-sockaddr-in))
         (s-port (sockint::sockaddr-in-port sockaddr))
         (s-addr (sockint::sockaddr-in-addr sockaddr))
         (fd (sockint::accept (socket-file-descriptor socket)
                              sockaddr
                              sockint::size-of-sockaddr-in)))
    (values
     (if (= fd -1)
         (socket-error "accept") 
      (let ((s (make-instance 'inet-socket :descriptor fd)))
        (ext:finalize s (lambda () (socket-close s)))))
     (vector
      (alien::deref s-addr 0)
      (alien::deref s-addr 1)
      (alien::deref s-addr 2)
      (alien::deref s-addr 3))
     (+ (* 256 (alien::deref s-port 0)) (alien::deref s-port 1)))))
#|
we can't test this in an automated fashion until we have the client code
to call it with

(defun test-foo ()
  (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp"))))
    (socket-bind s 9897 inet-address-any)
    (socket-listen s 5)
    (dotimes (i 5)
      (multiple-value-bind (s2 address port) (socket-accept s)
        (format t "Accepted connection from ~A:~A~%" address port)))))



|#

(defmethod socket-connect ((socket inet-socket) address port)
  (let* ((sockaddr (sockint::allocate-sockaddr-in))
         (s-port (sockint::sockaddr-in-port sockaddr))
         (s-addr (sockint::sockaddr-in-addr sockaddr)))
    (setf (sockint::sockaddr-in-family sockaddr) sockint::af-inet)
    (setf (alien::deref s-port 0) (ldb (byte 8 8) port))
    (setf (alien::deref s-port 1) (ldb (byte 8 0) port))
    (setf (alien::deref s-addr 0) (elt address 0))
    (setf (alien::deref s-addr 1) (elt address 1))
    (setf (alien::deref s-addr 2) (elt address 2))
    (setf (alien::deref s-addr 3) (elt address 3))

    (if (= (sockint::connect (socket-file-descriptor socket)
                             sockaddr
                             sockint::size-of-sockaddr-in)
           -1)
        (socket-error "connect") )))

(defmethod socket-peername ((socket inet-socket))
  ;;; XXX free sockaddr
  (let* ((sockaddr (sockint::allocate-sockaddr-in))
         (s-port (sockint::sockaddr-in-port sockaddr))
         (s-addr (sockint::sockaddr-in-addr sockaddr)))
    (if (= (sockint::getpeername (socket-file-descriptor socket)
                                 sockaddr
                                 sockint::size-of-sockaddr-in)
           -1)
        (socket-error "getpeername")
      (values
       (vector
        (alien::deref s-addr 0)
        (alien::deref s-addr 1)
        (alien::deref s-addr 2)
        (alien::deref s-addr 3))
       (+ (* 256 (alien::deref s-port 0)) (alien::deref s-port 1))))))

(defmethod socket-name ((socket inet-socket))
  ;;; XXX free sockaddr
  (let* ((sockaddr (sockint::allocate-sockaddr-in))
         (s-port (sockint::sockaddr-in-port sockaddr))
         (s-addr (sockint::sockaddr-in-addr sockaddr)))
    (if (= (sockint::getsockname (socket-file-descriptor socket)
                                 sockaddr
                                 sockint::size-of-sockaddr-in)
           -1)
        (socket-error "getsockname")
      (values
       (vector
        (alien::deref s-addr 0)
        (alien::deref s-addr 1)
        (alien::deref s-addr 2)
        (alien::deref s-addr 3))
       (+ (* 256 (alien::deref s-port 0)) (alien::deref s-port 1))))))

(defmethod socket-send ((socket inet-socket)   ))
(defmethod socket-receive ((socket inet-socket)   ))

