(defpackage "SOCKET-TEST"
  (:use "CL" "SOCKETS" "RT"))

(in-package :socket-test)

;;; a real address
(deftest make-inet-address
  (equalp (make-inet-address "127.0.0.1")  #(127 0 0 1))
  t)
;;; and an address with bit 8 set on some octets
(deftest make-inet-address2
  (equalp (make-inet-address "242.1.211.3")  #(242 1 211 3))
  t)

(deftest make-inet-socket
  ;; make a socket
  (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp"))))
    (and (> (socket-file-descriptor s) 1) t))
  t)

(deftest make-inet-socket-wrong
  ;; fail to make a socket: check correct error return.  There's no nice
  ;; way to check the condition stuff on its own, which is a shame
  (handler-case
   (make-inet-socket :stream (get-protocol-by-name "udp"))
   (protocol-not-supported-error (c) t)
   (:no-error nil))
  t)

(deftest make-inet-socket-check-collected
  ;; XXX make inet socket, GC several times, and check that the
  ;; file descriptor has been reclaimed
  'no-test-written-yet t)

(deftest inet-socket-bind
  (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp"))))
    ;; Given the functions we've got so far, if you can think of a
    ;; better way to make sure the bind succeeded than trying it
    ;; twice, let me know
    ;; 1974 has no special significance, unless you're the same age as me
    (ext:gc :full t) ;gc should clear out any old sockets bound to this port
    (or 
     (handler-case
      (socket-bind s (make-inet-address "127.0.0.1") 1974)
      (address-in-use-error () "Can't run this test: address in use. Collecting garbage may help.")
      nil)
     (handler-case
      (let ((s2 (make-inet-socket :stream (get-protocol-by-name "tcp"))))
        (socket-bind s2 (make-inet-address "127.0.0.1") 1974)
        nil)
      (sockets:address-in-use-error () t))))
  t)

(deftest simple-sockopt-test
  ;; test we can set SO_REUSEADDR on a socket and retrieve it, and in
  ;; the process that all the weird macros in sockopt happened right.
  (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp"))))
    (setf (sockopt-reuse-address s) t)
    (sockopt-reuse-address s))
  t)

;;; these require that the echo services are turned on in inetd

(deftest simple-tcp-client
  (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp")))
        (data (make-string 200)))
    (socket-connect s #(127 0 0 1) 7)
    (let ((stream (socket-make-stream s :input t :output t :buffering :none)))
      (format stream "here is some text")
      (let ((data (subseq data 0 (read-sequence data stream))))
        (format t "~&Got ~S back from TCP echo server~%" data)
        (> (length data) 0))))
  t)

(deftest simple-udp-client
  (let ((s (make-inet-socket :datagram (get-protocol-by-name "udp")))
        (data (make-string 200)))
    (socket-connect s #(127 0 0 1) 7)
    (let ((stream (socket-make-stream s :input t :output t :buffering :none)))
      (format stream "here is some text")
      (let ((data (subseq data 0 (read-sequence data stream))))
        (format t "~&Got ~S back from UDP echo server~%" data)
        (> (length data) 0))))
  t)

;;; these require that the internet (or, at least, my bit of it) is available

(deftest get-host-by-name
  (equalp (car (host-ent-addresses (get-host-by-name "loaclhost.telent.net")))
          #(212 113 30 134))
  t)

(deftest simple-http-client
  (handler-case
   (let ((s (make-inet-socket :stream (get-protocol-by-name "tcp")))
         (data (make-string 8192)))
     (socket-connect s #(212 113 30 134) 80)
     (let ((stream (socket-make-stream s :input t :output t :buffering :none)))
       (format stream "HEAD / HTTP/1.0~%~%")
       (let ((data (subseq data 0 (read-sequence data stream))))
         (princ data)
         (> (length data) 0))))
   (network-unreachable-error () 'network-unreachable))
  t)

