(in-package "SB-SYS")

;;;; This software is part of the SBCL system. See the README file for
;;;; more information.
;;;;
;;;; This software is derived from the CMU CL system, which was
;;;; written at Carnegie Mellon University and released into the
;;;; public domain. The software is in the public domain and is
;;;; provided with absolutely no warranty. See the COPYING and CREDITS
;;;; files for more information.


;;;; When stripped of all the junk for systems that don't have dlsym(),
;;;; this turns out to be fairly straightforward.
;;;;
;;;; LOAD-FOREIGN takes a .o file and a list of libraries.  It runs
;;;; ld(1) on them to produce a single file (in /tmp -- smart move, not)
;;;; and calls LOAD-OBJECT-FILE on that
;;;;
;;;; LOAD-OBJECT-FILE calls ENSURE-LISP-TABLE-OPENED, then dlopens the
;;;; file created above, and pushes the dlopen return onto *global-table*
;;;;
;;;; ENSURE-LISP-TABLE-OPENED does dlopen(NULL) (and push the answer)
;;;; unless *global-table* is non-NIL
;;;;
;;;; ALTERNATE-GET-GLOBAL-ADDRESS cdrs through *global-table* and
;;;; checks for the symbol in each element.  I (dan) suspect it does
;;;; more work than it needs to

;;;; Thing which is actually stopping it from working: we can't
;;;; implement LOAD-FOREIGN without RUN-PROGRAM.  LOAD-OBJECT-FILE
;;;; is useful anyway, though


;;; XXX there should really be a per-machine file full of symbolic constants.
;;; We could put the next three in it
(defconstant rtld-lazy 1  #+sb-doc  "lazy function call binding")
(defconstant rtld-now 2  #+sb-doc  "immediate function call binding")
(defconstant rtld-global #-irix #x100 #+irix 4
  #+sb-doc
  "If set the symbols of the loaded object and its dependencies are
   made visible as if the object were linked directly into the program")

(defvar *global-table* nil "list of SAPs corresponding to \"handles\"
of dynamic libraries as returned by calls to dlopen(3)")

;;; Dynamically loaded stuff isn't there upon restoring from a
;;; save--this is primarily for irix, which resolves tzname at
;;; runtime, resulting in *global-table* being set in the saved core
;;; image, resulting in havoc upon restart.

(pushnew (lambda () (setq *global-table* nil))
	 sb-int:*after-save-initializations*)

(sb-alien:def-alien-routine dlopen system-area-pointer
  (file sb-c-call:c-string) (mode sb-c-call:int))
(sb-alien:def-alien-routine dlsym system-area-pointer
  (lib system-area-pointer)
  (name sb-c-call:c-string))
;;;(sb-alien:def-alien-routine dlclose void (lib system-area-pointer))
(sb-alien:def-alien-routine dlerror sb-c-call:c-string)

;;; Ensure we've opened our own binary so can resolve global variables
;;; in the lisp image that come from libraries. This used to happen
;;; only in alternate-get-global-address, and only if no libraries
;;; were dlopened already, but that didn't work if something was
;;; dlopened before any problem global vars were used. So now we do
;;; this in any function that can add to the global-table, as well as
;;; in alternate-get-global-address.

;;; (The implication of the preceding seems to be that there exists
;;; some system where you can't dlopen(NULL) if you've previously done
;;; dlopen("some filename").  The dlopen(3) man page makes no mention
;;; of this, fwiw  -dan

(defun ensure-lisp-table-opened ()
  (unless *global-table*
    ;; Prevent recursive call if dlopen isn't defined.
    (setf *global-table* (int-sap 0))
    (setf *global-table* (list (dlopen nil rtld-lazy)))
    (when (zerop (sb-sys:sap-int (car *global-table*)))
      (error "can't open global symbol table: ~S" (dlerror)))))

(defun load-object-file (file)
  (ensure-lisp-table-opened)
  ;; rtld global: symbols will be accessible to other objects opened later
  ;; rtld now: dlopen will fail if not all symbols are defined.
  (let ((sap (dlopen file (logior rtld-now rtld-global))))
       (if (zerop (sap-int sap))
	   (error "Can't open object ~S: ~S" file (dlerror))
	   (pushnew sap *global-table* :test #'sap=))))

;;; this is called from sb!impl::foreign-symbol-address-aux if it
;;; can't find a suitable symbol in sb!impl::*foreign-symbols*.  I
;;; can't actually see anywhere that *foreign-symbols* is set, so I
;;; suspect it gets called quite a lot.

(defun alternate-get-global-address (symbol)
  (ensure-lisp-table-opened)
  ;; Find the symbol in any of the loaded objects. Search in reverse order of
  ;; loading, so that later loadings take precedence.
  ;; (Given that all the libraries were loaded with RTLD_GLOBAL, I suspect that
  ;; this could be reduced to a single call to dlsym()   -dan, 20000703)
  (let ((result 0))
       (do ((table *global-table* (cdr table)))
	   ((or (null (car table)) (not (zerop result))))
	   (setq result (sap-int (dlsym (car table) symbol))))
       (values result)))

#|

;;; This would - if it were uncommented and fixed - implement the
;;; foreign code interface documented in the CMUCL manual.

(defun pick-temporary-file-name (&optional (base "/tmp/tmp~D~C"))
  ;; SECURITY FIXME we're asking for a /tmp/ race here
  (let ((code (char-code #\A)))
    (loop
      (let ((name (format nil base (sb-unix:unix-getpid) (code-char code))))
	(multiple-value-bind (fd errno)
	    (sb-unix:unix-open name
			       (logior sb-unix:o_wronly
				       sb-unix:o_creat
				       sb-unix:o_excl)
			       #o666)
	  (cond ((not (null fd))
		 (sb-unix:unix-close fd)
		 (return name))
		((not (= errno sb-unix:eexist))
		 (error "could not create temporary file ~S: ~A"
			name
			(sb-unix:get-unix-error-msg errno)))
		;; KLUDGE: depends on ASCII character ordering -- WHN 20000128
		((= code (char-code #\Z))
		 (setf code (char-code #\a)))
		((= code (char-code #\z))
		 (return nil))
		(t
		 (incf code))))))))

(defvar *dso-linker*
  #+solaris "/usr/ccs/bin/ld"
  #+(or linux irix) "/usr/bin/ld")

(defun load-foreign (files &key
			   (libraries '("-lc"))
			   (base-file nil)
			   ;; Note: Now that SBCL has no *ENVIRONMENT-LIST*
			   ;; variable, if this code is ever restored,
			   ;; the default should be taken from the alien
			   ;; "environ" variable.
			   (env sb-ext:*environment-list*))
  #+sb-doc
  "Load-foreign loads a list of C object files into a running Lisp. The files
  argument should be a single file or a list of files. The files may be
  specified as namestrings or as pathnames. The libraries argument should be a
  list of library files as would be specified to ld. They will be searched in
  the order given. The default is just \"-lc\", i.e., the C library. The
  base-file argument is used to specify a file to use as the starting place for
  defined symbols. The default is the C start up code for Lisp. The env
  argument is the Unix environment variable definitions for the invocation of
  the linker. The default is the environment passed to Lisp."
  ;; Note: dlopen remembers the name of an object, when dlopenin
  ;; the same name twice, the old objects is reused.
  (declare (ignore base-file))
  (let ((output-file (pick-temporary-file-name
		      (concatenate 'string "/tmp/~D~C" (string (gensym)))))
	(error-output (make-string-output-stream)))

    (format t ";;; Running ~A...~%" *dso-linker*)
    (force-output)
    (let ((proc (sb-ext:run-program
		 *dso-linker*
		 (list*
			#+(or solaris linux) "-G" #+irix "-shared"
			"-o"
			output-file
			(append (mapcar #'(lambda (name)
					    (unix-namestring name nil))
					(if (atom files)
					    (list files)
					    files))
				libraries))
		 :env env
		 :input nil
		 :output error-output
		 :error :output)))
      (unless proc
       (error "could not run ~A" *dso-linker*))
      (unless (zerop (sb-ext:process-exit-code proc))
	(sb-sys:serve-all-events 0)
	(error "~A failed:~%~A" *dso-linker*
	       (get-output-stream-string error-output)))
      (load-object-file output-file)
      (sb-unix:unix-unlink output-file)
      ))
  (format t ";;; Done.~%"))
|#


