(in-package "SOCKETS")

;;;; Methods, classes, functions for sockets.  Protocol-specific stuff
;;;; is deferred to inet.lisp, unix.lisp, etc

(eval-when (:load-toplevel :compile-toplevel :execute)
(defclass socket ()
  ((file-descriptor :initarg :descriptor :reader
                    socket-file-descriptor)
     (stream))))
  
(defmethod print-object ((object socket) stream)
  (print-unreadable-object (object stream :type t :identity t)
                           (princ "descriptor " stream)
                           (princ (slot-value object 'file-descriptor) stream)))

(defun get-protocol-by-name (name)      ;exported
  "Returns the network protocol number associated with the string NAME,
using getprotobyname(2) which typically looks in NIS or /etc/protocols"
  ;; for extra brownie points, could return canonical protocol name
  ;; and aliases as extra values
  (let ((ent (sockint::getprotobyname name)))
    (sockint::protoent-proto ent)))



(defgeneric socket-bind ((s socket) &rest address))

;;; Methods common to sockets in all domains

(defmethod socket-listen ((socket socket) backlog)
  "Mark SOCKET as willing to accept incoming connections.  BACKLOG
defines the maximum length that the queue of pending connections may
grow to before new connection attempts are refused.  See also listen(2)"
  (let ((r (sockint::listen (socket-file-descriptor socket) backlog)))
    (if (= r -1)
        (socket-error "listen"))))

(defmethod socket-close ((socket socket))
  "Close SOCKET.  May throw any kind of error that write(2) would have
thrown.  If SOCKET-MAKE-STREAM has been called, calls CLOSE on that
stream instead"
  ;; the close(2) manual page has all kinds of warning about not
  ;; checking the return value of close, on the grounds that an
  ;; earlier write(2) might have returned successfully w/o actually
  ;; writing the stuff to disk.  It then goes on to define the only
  ;; possible error return as EBADF (fd isn't a valid open file
  ;; descriptor).  Presumably this is an oversight and we could also
  ;; get anything that write(2) would have given us.

  ;; What we do: we catch EBADF.  It should only ever happen if
  ;; (a) someone's closed the socket already (stream closing seems
  ;; to have this effect) or (b) the caller is messing around with
  ;; socket internals.  That's not supported, dude
  
  (unless (slot-boundp socket 'stream)
    (handler-case
     (if (= (sockint::close (socket-file-descriptor socket)) -1)
         (socket-error "close"))
     (bad-file-descriptor-error () nil)
     (:no-error nil))))

(defmethod socket-make-stream ((socket socket)  &rest args)
  "Find or create a STREAM that can be used for IO on SOCKET (which
must be connected).  ARGS are passed onto SB-SYS:MAKE-FD-STREAM."
  (let ((stream
         (or (and (slot-boundp socket 'stream) (slot-value socket 'stream))
             (apply #'sb-sys:make-fd-stream (socket-file-descriptor socket)
                    :auto-close nil args))))
    (setf (slot-value socket 'stream) stream)))



;;; Error handling

;;; is CONDITION the right condition to subclass this from?  ERROR
;;; probably isn't; Some of the `errors' are going to be things like
;;; EAGAIN which is not really all that grave

;;; hairier is that when people get their STREAM and use the stuff in the
;;; rest of CMUCL to read/write it, they'll never see our neat conditions
;;; and will be dependent on whatever error checking goes on in CMUCL proper.
;;; I don't think there's much we can do about this other than be aware of it

(define-condition socket-error (condition)
  ((errno :initform nil
          :initarg :errno  
          :reader socket-error-errno) 
   (symbol :initform nil :initarg :symbol :reader socket-error-symbol)
   (syscall  :initform "outer space" :initarg :syscall :reader socket-error-syscall))
  (:report (lambda (c s)
             (let ((num (socket-error-errno c)))
               (format s "Socket error in \"~A\": ~A (~A)"
                       (socket-error-syscall c)
                       (or (socket-error-symbol c) (socket-error-errno c))
                       (sb-unix:get-unix-error-msg num))))))

;;; watch out for slightly hacky symbol punning: we use both the value
;;; and the symbol-name of sockint::efoo

(defmacro define-socket-condition (symbol name)
  `(progn
     (define-condition ,name (socket-error)
       ((symbol :reader socket-error-symbol :initform (quote ,symbol))))
     (push (cons ,symbol (quote ,name)) *conditions-for-errno*)))

(defparameter *conditions-for-errno* nil)
;;; this needs the rest of the list adding to it, really.  They also
;;; need
;;; - conditions to be exported in the DEFPACKAGE form
;;; - symbols to be added to constants.ccon
;;; I haven't yet thought of a non-kludgey way of keeping all this in
;;; the same place
(define-socket-condition sockint::EADDRINUSE address-in-use-error)
(define-socket-condition sockint::EAGAIN interrupted-error)
(define-socket-condition sockint::EBADF bad-file-descriptor-error)
(define-socket-condition sockint::ECONNREFUSED connection-refused-error)
(define-socket-condition sockint::EINTR interrupted-error)
(define-socket-condition sockint::EINVAL invalid-argument-error)
(define-socket-condition sockint::ENOBUFS no-buffers-error)
(define-socket-condition sockint::ENOMEM out-of-memory-error)
(define-socket-condition sockint::EOPNOTSUPP operation-not-supported-error)
(define-socket-condition sockint::EPERM operation-not-permitted-error)
(define-socket-condition sockint::EPROTONOSUPPORT protocol-not-supported-error)
(define-socket-condition sockint::ENETUNREACH network-unreachable-error)



(defun condition-for-errno (err)
  (or (cdr (assoc err *conditions-for-errno* :test #'eql)) 'socket-error))
      
(defun socket-error (where)
  (sb-unix::unix-get-errno)
  (let ((condition (condition-for-errno sb-unix:unix-errno)))
    (error condition :errno sb-unix:unix-errno  :syscall where)))



