(in-package :telentweb)

(defun read-sentence (stream)
  "Read until end of sentence"
  (with-output-to-string (o)
    (let (c in-tag)
      (loop 
       (setf c (read-char stream nil nil))
       (when (and (not in-tag)
		  (or (not c)
		      (and (member c '(#\. #\? #\!))
			   (not (alphanumericp 
				 (peek-char nil stream nil #\Return))))
		      (and (not (graphic-char-p c))
			   (not (graphic-char-p
				 (peek-char nil stream nil #\Return))))))
	 (return))
       (if (and (not in-tag) (eql c #\<)) (setf in-tag t))
       (unless in-tag (princ c o))
       (if (and in-tag (eql c #\>)) (setf in-tag nil))))))

(defclass diary-handler (araneida:handler)
  ((data-directory :initarg :data-directory :accessor diary-data-directory)))

(defclass diary-entry ()
  ((date :accessor entry-date :initarg :date)))

(defmethod entry-pathname ((handler diary-handler) (entry diary-entry))
  (let ((*default-pathname-defaults* (diary-data-directory handler)))
    (merge-pathnames
     (make-pathname :name (princ-to-string (entry-date entry))
		    :type "html"))))

(defmethod print-entry ((entry diary-entry) request)
  (date:with-decoding ((entry-date entry) 0)
    (let ((stream (request-stream request))
	  (base (urlstring (request-base-url request)))
	  (offset (+ date:second (* 60 (+ date:minute (* 60 date:hour))))))
      (with-open-file (i (entry-pathname (caar (request-handled-by request))
					 entry))
	(format stream "<h2><a name=\"~A.~A\">~A</a> (<a href=\"~A~A/~A/#~A.~A\">#</a>)</h2>~%~%"
		date:day-of-month offset
		(date:universal-time-to-rfc-date (entry-date entry))
		base date:year date:month date:day-of-month
		offset)
	(araneida::copy-stream i stream)))))

(defmethod rss-for-entry ((entry diary-entry) request)
  (date:with-decoding ((entry-date entry) 0)
    (with-open-file (i (entry-pathname 
			(caar (request-handled-by request))
			entry))
      (let ((base (urlstring (request-base-url request)))
	    (text (make-string (file-length i)))
	    (offset (+ date:second (* 60 (+ date:minute (* 60 date:hour)))))
	    (fpos (file-position i))
	    (description (read-sentence i)))
	(file-position i fpos)
	(read-sequence text i)
	`("item"
	  ()
	  ("title" () ,description)
	  ("pubDate" () ,(date:universal-time-to-rfc-date (entry-date entry)))
	  ("author" () "dan@telent.net")
	  ("link" () 
		  ,(format nil "~A~A/~A/#~A.~A"
			   base date:year date:month date:day-of-month offset))
	  ("description" 
	   ()
	   ,text))))))

(defun entries-for (diary year month)
  ;; and a few days either side for continuity
  (let* ((month-start
	  (encode-universal-time 0 0 0 1 month year))
	 (month-end (+ month-start (* 86400 31)))
	 (*default-pathname-defaults* (diary-data-directory diary))
	 (ret nil) (next-month-p nil) (previous-month-p nil))
    (dolist (file (directory "*.html" ))
      (let ((date
	     (or (parse-integer (pathname-name file) :junk-allowed t) 0)))
	(if (<= (- month-start (* 86400 5))
		date
		(+ month-end (* 86400 5)))
	    (push (make-instance 'diary-entry :date date) ret))
	(if (> date month-end ) (setf next-month-p t))
	(if (< date month-start ) (setf previous-month-p t))))
    (values ret next-month-p previous-month-p)))

(defun recent-entries (diary n)
  (let ((*default-pathname-defaults* (diary-data-directory diary)))
    (mapcar (lambda (d) (make-instance 'diary-entry :date d))
	    (subseq (sort (mapcar 
			   (lambda (x)
			     (cliki::integer-for (pathname-name x)))
			   (directory "*.html")) 
			  #'>) 0 n))))


(defmethod handle-request-response ((handler diary-handler) method request)
  (when (string= (pathname-type (request-path-info request)) "rss") ;; ick
    (rss-entries handler request)
    (signal 'response-sent))
  (let* ((base (request-base-url request))
	 (*default-pathname-defaults* (diary-data-directory handler))
	 (now (multiple-value-list (get-decoded-time)))
	 (stream (request-stream request))
	 (bits (split-sequence:split-sequence #\/ (request-path-info request)))
	 (year (or (parse-integer (or (first bits) "") :junk-allowed t)
		   (sixth now)))
	 (month (or (parse-integer (or (second bits) "")
				   :junk-allowed t)
		    (fifth now))))
    (multiple-value-bind (entries next-p prev-p)
	(entries-for handler year month)
      (let ((last-modified
	     (reduce #'max entries
		     :key
		     (lambda (x) (file-write-date (entry-pathname handler x)))
		     :initial-value 0)))
	(if (zerop last-modified) (setf last-modified (get-universal-time)))
	(format t "~A ~A~%" last-modified (request-if-modified-since request))
	(request-send-headers request
			      :conditional t
			      :last-modified last-modified))
      (with-open-file (head "HEADER.inc")
	(araneida::copy-stream head stream))
      (unless (and (= year (sixth now)) (= month (fifth now)))
	(format stream "<ul><li>Upward (metaphorically) to <a href=\"~A\">today</a> or something in that vicinity</ul>~%"
		(urlstring base)))
      (when next-p
	(let* ((m (1+ (mod month 12)))
	       (y (if (< m month) (1+ year) year)))
	  (format stream "<ul><li>Onward to <a href=\"~A~A/~A/\">~/date:monthname/ ~A</a></ul>~%"
		  (urlstring base) y m m y)))
      (dolist (e entries)
	(print-entry  e request))
      (when prev-p
	(let* ((m (1+ (mod (+ 10 month) 12)))
	       (y (if (> m month) (1- year) year)))
	  (format stream "<ul><li>Backward to <a href=\"~A~A/~A/\" >~/date:monthname/ ~A</a></ul>~%"
		  (urlstring base)  y m m y)))
      (with-open-file (foot "FOOTER.inc")
	(araneida::copy-stream foot stream))
      ))
  t)

(defun rss-entries (handler request)
  (let* ((*default-pathname-defaults* (diary-data-directory handler))
	 (stream (request-stream request))
	 (entries (recent-entries handler 15)))
    (let ((last-modified
	   (reduce #'max entries
		   :key
		   (lambda (x) (file-write-date (entry-pathname handler x)))
		   :initial-value 0)))
	(request-send-headers request
			      :expires (+ (get-universal-time) 600)
			      :conditional t
			      :content-type "application/rss+xml"
			      :last-modified last-modified))
    (format stream "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>~%")
    (xmls:write-xml
     `("rss" (("version" "2.0"))
	     ("channel"
	      ()
	      ("title" 
	       () "Daniel Barlow's diary")
	      ("description" 
	       () "More than you needed to know: SBCL internals and skin loss")
	      ("link" () "http://ww.telent.net/diary/")
	      ,@(mapcar (lambda (x) (rss-for-entry x request)) entries)))
     stream)
    t))
