;;;;
;;;; This code was written by Daniel Barlow <dan@telent.net>, closely
;;;; based on code from Christophe Rhodes <csr21@cam.ac.uk>.  It's all
;;;; in the public domain.

;;; Class mail-output-stream.
;;;
;;; The mail-output-stream is a class of stream (therefore depending
;;; on the Gray Stream extension) that sends anything output to it
;;; by email.
;;;

(defpackage "SMTP"
  (:use "COMMON-LISP"
	#+cmu "EXTENSIONS"
	#+sbcl "SB-GRAY"
	#+sbcl "SB-BSD-SOCKETS"
	#+allegro "STREAM")
  (:export "MAIL-OUTPUT-STREAM" "MAIL-ERROR" "WITH-MAIL-OUTPUT-STREAM"
	   "ERROR-STREAM" "ERROR-COMMAND" "ERROR-RESPONSE" ))
(in-package "SMTP")

(defclass mail-output-stream (fundamental-character-output-stream)
  ;; user-settable
  ((sender :accessor sender :initform  nil :initarg :sender)
   (recipients :accessor recipients :initform nil :initarg :recipients)
   (server-host :initform "localhost"
		:initarg :server-host :reader server-host)
   (server-port :initform 25
		:initarg :server-port :reader server-port)
   (delay-recipient-errors
    :initform nil :reader delay-recipient-errors
    :initarg :delay-recipient-errors
    :documentation "If non-NIL, errors from RCPT commands are not reported until after mail body has been sent.  This will cause bounce messages")
   ;; internal
   (line-column :initform 0 :accessor stream-line-column)
   (rcpt-responses :accessor rcpt-responses)
   (real-stream :initarg :stream :initform nil
		:accessor mail-output-stream-stream)
   (delayed-error :initform nil :accessor delayed-error)
   (socket-stream :accessor socket-stream :initform nil)))

(defmethod stream-start-line-p ((stream mail-output-stream))
  (eql (stream-line-column stream) 0))

(defmacro with-mail-output-stream ((stream &rest args) &body body)
  `(let ((,stream (make-instance 'mail-output-stream ,@args)))
    (unwind-protect
	 ,@body
      (close ,stream))))

;;; tacky, no?  We don't know what line ending the lisp implementation
;;; conventionally uses, so we'd better handle it ourselves.
(defvar *crlf* #.(coerce (list (code-char 13) (code-char 10)) 'string))

;;; replies are all prefixed with a numeric code.  The last line of
;;; a reply is (code SPACE text) ; preceding lines (if any) are
;;; (code "-" text)
(defun read-response (stream)
  "Read a complete SMTP response from STREAM, returning list of
numeric-code text"
  (let* ((code "500")
         (text
          (with-output-to-string (o)
            (loop for l =  (read-line stream nil nil)
                  while l
                  for sep = (position-if-not #'digit-char-p l)
                  do (setf code (subseq l 0 sep))
                  do (let ((cr (position (code-char 13) l :start sep)))
		       (write-sequence l o :start (and sep (1+ sep)) :end cr)
		       (when (eql (elt l sep) #\-) (terpri o)))
                  until (or (not l) (eql (elt l 3) #\Space)))))
	 (code (parse-integer code))
	 (keyword
	  (cond ((<= 200 code 399) :ok)
		((<= 400 code 499) :temporary-error)
		((<= 500 code 599) :permanent-error))))
    (list keyword code text)))

(defun ok-p (response) (eql (car response) :ok))
(defun error-p (response)  (not (ok-p response)))

(defun open-tcp-connection (server port)
  "Open connection to SERVER on PORT, returning a stream or NIL on failure"
  (let ((address (car (host-ent-addresses (get-host-by-name server))))
        (socket (make-instance 'inet-socket :type :stream :protocol :tcp )))
    (socket-connect socket address port)
    (socket-make-stream socket :input t :output t :buffering :none)))

(defmacro send-and-test (stream error &optional format &rest args)
  (let ((r (gensym "RESPONSE"))
	(s (gensym "SOCKET")))
    `(let ((,s (socket-stream ,stream)))
      ,(if format `(format ,s ,format ,@args))
      (let ((,r (read-response ,s)))
	(when (error-p ,r)
	  (error 'mail-error :command ,error
		 :responses (rcpt-responses ,stream)
		 :response ,r :stream ,stream))))))

(defmethod initialize-instance :after ((stream mail-output-stream) &key)
  (let ((socket
	 (open-tcp-connection (server-host stream) (server-port stream)))
	(some-success-p nil)
	(some-failure-p nil))
    (setf (socket-stream stream) socket
	  (rcpt-responses stream)
	  (pairlis (recipients stream)
		   (make-list (length (recipients stream))
			      :initial-element
			      '(:temporary-error 500 "Not attempted"))))
    (if (null (mail-output-stream-stream stream))
	(setf (mail-output-stream-stream stream) socket)
	(setf (mail-output-stream-stream stream)
	      (make-broadcast-stream (mail-output-stream-stream stream)
				     socket)))
    (unwind-protect
	 (progn
	   (send-and-test stream nil)
	   (send-and-test stream :helo "HELO ~A~A" "localhost" *crlf*)
	   (send-and-test stream :from
			  "MAIL FROM:<~A>~A" (sender stream) *crlf*)
	   (dolist (i (rcpt-responses stream))
	     (let ((r (progn (format socket "RCPT TO:<~A>~A" (car i) *crlf*)
			     (read-response socket))))
	       (rplacd i r)
	       (if (ok-p r)
		   (setf some-success-p r)
		   (setf some-failure-p r))))
	   (when some-failure-p
	     (let ((c (make-condition 'mail-error :command :rcpt
				      :responses (rcpt-responses stream)
				      :stream stream
				      :response some-failure-p)))
	       (if (and some-success-p
			(delay-recipient-errors stream))
		   ;; squirrel the condition away for later
		   (setf (delayed-error stream) c)
		   (error c))))
	   (send-and-test stream :data "DATA~A" *crlf*))
      (unless some-success-p
	(close socket)
	(setf (socket-stream stream) nil)))))

(defmethod print-object ((object mail-output-stream) stream)
  (print-unreadable-object (object stream :type t :identity t)
    (format stream "socket ~A" (recipients object) )))


;;; all this mess serves two purposes: (1) to translate Lisp's newline
;;; into CRLF as required by SMTP (yes, some implementations may have
;;; neat stuff to do this already.  Dunno which), (2) lines that
;;; begin with a #\. must have the dot doubled to avoid having it
;;; signal the end of the message
(defmethod stream-write-string ((stream mail-output-stream) string
				&optional (start 0) (end (length string)))
  (unless end (setf end (length string)))
  (let ((o (mail-output-stream-stream stream)))
    (loop 
     (when (>= start end) (return))
     (when (and (stream-start-line-p stream)
		(eql (elt string start) #\.)) (princ #\. o))
     (let ((newline (position #\Newline string :start start :end end)))
       (cond (newline
	      (write-sequence string o :start start :end newline)
	      (princ (code-char 13) o)
	      (princ (code-char 10) o)
	      (setf (stream-line-column stream) 0)
	      (setf start (1+ newline)))
	     (t
	      (write-sequence string o :start start :end end)
	      (incf (stream-line-column stream) (- end start ))
	      (setf start end)))))))

;;; having done all that once ... let's just do write-char the slow way
(defmethod stream-write-char ((stream mail-output-stream) character)
  (stream-write-string stream (string character)))

(defmethod stream-finish-output ((stream mail-output-stream))
  (finish-output (mail-output-stream-stream stream)))

(defmethod stream-force-output ((stream mail-output-stream))
  (force-output (mail-output-stream-stream stream)))

(defmethod stream-clear-output ((stream mail-output-stream))
  (clear-output (mail-output-stream-stream stream)))

(defmethod close ((stream mail-output-stream) &key abort)
  (when (mail-output-stream-stream stream)
    (close (mail-output-stream-stream stream) :abort abort))
  (when (socket-stream stream) (close (socket-stream stream) :abort abort))
  (when (delayed-error stream) (error (delayed-error stream))))

(define-condition mail-error (error)
  ((stream :reader error-stream :initarg :stream)
   (command :reader error-command :initarg :command)
   (response :reader error-response :initarg :response)
   (recipients :reader error-recipients :initarg :recipients))
  (:report (lambda (condition stream)
	     (format stream "SMTP error on ~A during ~A: ~A"
		     (error-stream condition)
		     (error-command condition)
		     (error-response condition)))))


(defmethod close :before ((stream mail-output-stream)
			  &key abort &allow-other-keys)
  (unless abort
    (let ((socket (socket-stream stream)))
      (when socket
	(unless (stream-start-line-p stream)
	  (princ *crlf* socket))
	(princ #\. socket)
	(send-and-test stream :data *crlf*)
	(send-and-test stream :quit "QUIT~A" *crlf*)))))
    
(eval-when (:load-toplevel :execute)
  (pushnew :mail-streams *features*))
